import networkx as nx
from networkx.algorithms.components.connected import connected_components


def to_graph(elements):
    """.

    Parameters
    ----------
    elements :
        elements
    """
    G = nx.Graph()
    for part in elements:
        # each sublist is a bunch of nodes
        G.add_nodes_from(part)
        # it also imlies a number of edges:
        G.add_edges_from(to_edges(part))
    return G


def to_edges(elements):
    """Treat elements as graph and return it's edges.

    Parameters
    ----------
    elements
        Container with elements

    Examples
    --------
    >>> to_edges(['a','b','c','d'])
    [(a,b), (b,c),(c,d)]
    """
    it = iter(elements)
    last = next(it)

    for current in it:
        yield last, current
        last = current


def connected_components_from_container(elements):
    # https://stackoverflow.com/a/4843408
    G = to_graph(elements)
    return connected_components(G)



"""base classes for datasets."""
import os
import pathlib
from typing import List, Tuple
from zipfile import ZipFile

import wget


class Dataset:
    """Base class for dataset classes."""

    DS_NAME = ""


class RemoteDataset(Dataset):
    """Base class for dataset with remote files."""

    def __init__(self, download_urls: List[Tuple[str, str]], data_folder: str = None):
        """Initialize a dataset object.

        Parameters
        ----------
        download_urls : List[Tuples(str, str)]
            List of tuples with (url, target_file_name)
        data_folder : str
            folder path where raw data is stored.
            If NONE, stores in "forayer_root/data/self.__class__.DS_NAME"
        """
        # get root of project
        project_root = pathlib.Path(__file__).parent.parent.parent.resolve()
        self.data_folder = (
            data_folder
            if data_folder is not None
            else os.path.join(project_root, "data", self.__class__.DS_NAME)
        )
        self.download_urls = download_urls

    def download(self, check_files=False) -> bool:
        """Download files and create download dir if necessary.

        check_files: bool
            Instead of checking for data_folder check individual files.

        Returns
        -------
        bool
            True if download was needed.
        """
        if not os.path.exists(self.data_folder):
            print(
                f"Downloading {self.__class__.__name__} datasets to {self.data_folder}"
            )
            os.makedirs(self.data_folder)
            for url, target in self.download_urls:
                target_file_path = os.path.join(self.data_folder, target)
                print(f"Downloading {url} to {target_file_path}")
                wget.download(url, target_file_path)
            return True
        elif check_files:
            at_least_one = False
            for url, target in self.download_urls:
                target_file_path = os.path.join(self.data_folder, target)
                if not os.path.exists(target_file_path):
                    at_least_one = True
                    print(f"Downloading {url} to {target_file_path}")
                    wget.download(url, target_file_path)
            if at_least_one:
                return True
        return False


class ZipDataset(RemoteDataset):
    """Base class for datasets that have remote zipped raw files."""

    def __init__(self, data_folder: str, download_urls: List[Tuple[str, str]]):
        """Initialize a dataset object.

        Downloads and unzips necessary data

        Parameters
        ----------
        download_urls : List[Tuples(str, str)]
            List of tuples with (url, target_file_name)
        data_folder : str
            folder path where raw data is stored
        """
        super(ZipDataset, self).__init__(download_urls, data_folder)

    def download_and_unzip(self) -> bool:
        """Download and unzip files to download folder if needed.

        Returns
        -------
        bool
            True if download was needed.
        """
        if self.download():
            # get zip file paths
            for _, zip_name in self.download_urls:
                zip_dir = os.path.join(self.data_folder, zip_name)
                with ZipFile(zip_dir, "r") as zip_obj:
                    zip_obj.extractall(self.data_folder)
                # cleanup
                os.remove(zip_dir)
            return True
        return False



import sys
import warnings


def warning_formatter(msg, category, filename, lineno, line=None):
    """Format warning to only print filename, linenumber and message.

    Parameters
    ----------
    msg
        warning message
    category
        warning category
    filename
        filename of file where warning was raised
    lineno
        linenumber where warning was raised
    line
        line containing warning

    Returns
    -------
    str
        formatted warning message
    """
    return f"{filename}:L{lineno}: {msg}\n"


warnings.formatwarning = warning_formatter


if sys.version_info[1] > 7:
    from importlib.metadata import version  # pragma: no cover
else:
    from importlib_metadata import version  # pragma: no cover

__version__ = version(__package__)



"""module containing knowledge graph class."""
from __future__ import annotations

import random
import warnings
from collections import defaultdict
from itertools import chain
from typing import Any, Dict, Iterable, List, Set, Union

from forayer.transformation.word_embedding import AttributeVectorizer
from forayer.utils.dict_help import dict_merge, nested_ddict2dict
from forayer.utils.random_help import random_generator
from rdflib import Graph, Literal, URIRef
from tqdm import tqdm


class KG:
    """KG class holding entities and their attributes and relations between entities."""

    def __init__(
        self,
        entities: Dict[str, Dict[Any, Any]],
        rel: Dict[str, Dict[str, Any]] = None,
        name: str = None,
    ):
        """Initialize a KG object.

        Parameters
        ----------
        entities : Dict[str, Dict[Any, Any]]
            entity information with entity ids as keys and a attribute dictionaries as values
            attribute dictionaries have attribute id as key and attribute value as dict value
        rel : Dict[str, Dict[str, Any]]
            relation triples with one entity as key, value is dict with other entity as key
            and relation id as value
        name : str, optional
            name of the kg, default is None

        Examples
        --------
        >>> entities = {
            "e1": {"a1": "first entity", "a2": 123},
            "e2": {"a1": "second ent"},
            "e3": {"a2": 124},
        }
        >>> relations = {"e1": {"e3": "somerelation"}}
        >>> kg = KG(entities, relations, "mykg")
        """
        self.entities = entities
        self.rel = rel if rel is not None else {}
        inv_rel = {}
        if rel is not None:
            for left_ent, right_ent_rel in rel.items():
                for right_ent, _ in right_ent_rel.items():
                    if right_ent not in inv_rel:
                        inv_rel[right_ent] = set()
                    inv_rel[right_ent].add(left_ent)
        self._inv_rel = inv_rel
        self.name = name

    def __eq__(self, other):
        if isinstance(other, KG):
            return (
                self.entities == other.entities
                and self.rel == other.rel
                and self.name == other.name
            )
        return False

    def search(self, query, attr=None, exact=False):
        """Search for entities with specific attribute value.

        Parameters
        ----------
        query
            attribute value that is searched for
        attr : Union[str,List]
            only look in specific attribute(s)
        exact : bool
            if True only consider exact matches

        Returns
        -------
        result: Dict[str, Dict[str, Any]]
            Entites that have attribute values that match the query.

        Examples
        --------
        >>> from forayer.knowledge_graph import KG
        >>> entities = {
                "e1": {"a1": "first entity", "a2": 123},
                "e2": {"a1": "second ent"},
                "e3": {"a2": 124},
            }
        >>> kg = KG(entities)
        >>> kg.search("first")
        {'e1': {'a1': 'first entity', 'a2': 123}}
        >>> kg.search("first", exact=True)
        {}
        >>> kg.search("first", attr="a2")
        {}
        """
        query = str(query)
        if attr is not None and not isinstance(attr, list):
            attr = [attr]
        result = {}
        for ent_id, attr_dict in self.entities.items():
            if attr is not None:
                filtered_attr_dict = {k: v for k, v in attr_dict.items() if k in attr}
            else:
                filtered_attr_dict = attr_dict
            if exact:
                if query in filtered_attr_dict.values():
                    result[ent_id] = attr_dict
            else:
                for _, value in filtered_attr_dict.items():
                    if query in str(value):
                        result[ent_id] = attr_dict
        return result

    def with_attr(self, attr: str):
        """Search for entities with specific attribute.

        Parameters
        ----------
        attr: str
            Attribute name.

        Returns
        -------
        result: Dict[str, Dict[str, Any]]
            Entites that have the attribute.

        Examples
        --------
        >>> from forayer.knowledge_graph import KG
        >>> entities = {
                "e1": {"a1": "first entity", "a2": 123},
                "e2": {"a1": "second ent"},
                "e3": {"a2": 124},
            }
        >>> kg = KG(entities)
        >>> kg.with_attr("a1")
        {'e1': {'a1': 'first entity', 'a2': 123}, "e2": {"a1": "second ent"}}
        """
        return {
            ent_id: attr_dict
            for ent_id, attr_dict in self.entities.items()
            if attr in attr_dict
        }

    def subgraph(self, wanted: Iterable[str]):
        """Return a subgraph with only wanted entities.

        Creates a subgraph with the wanted entities. Contains only relationships
        between wanted entities. Entities without attributes (possibly not contained
        in self.entities) and relationships that point outside the subgraph are added
        as entities without attributes to the result KG's entities.

        Parameters
        ----------
        wanted: Iterable[str]
            Ids of wanted entities.

        Returns
        -------
        KG
            subgraph with only wanted entities

        Examples
        --------
        >>> from forayer.knowledge_graph import KG
        >>> entities = {"e1": {"a": 1}, {"e2": {"a": 3}}
        >>> rel = {"e1": {"e2": "rel", "e3": "rel"}}
        >>> kg = KG(entities,rel)
        >>> kg.subgraph(["e1","e3"])
        KG(entities={'e1': {'a': 1}, 'e3': {}}, rel={'e1': {'e3': 'rel'}}, name=None)
        """
        wanted_entities = {
            ent_id: attr_dict
            for ent_id, attr_dict in self.entities.items()
            if ent_id in wanted
        }
        wanted_rel = defaultdict(dict)
        entities_in_rel = set()
        for ent_id, right_rel_dict in self.rel.items():
            if ent_id in wanted:
                for right_ent_id, rel_dict in right_rel_dict.items():
                    if right_ent_id in wanted:
                        entities_in_rel.add(ent_id)
                        entities_in_rel.add(right_ent_id)
                        wanted_rel[ent_id][right_ent_id] = rel_dict
        # add entities without attributes, that only show up in relationships
        # that point outside the subgraph and therefore were missed
        for w in wanted:
            if (w not in wanted_entities and w not in wanted_rel) and (
                w in self.entities or w in self.rel or w in self._inv_rel
            ):
                wanted_entities[w] = {}
        return KG(
            entities=wanted_entities, rel=nested_ddict2dict(wanted_rel), name=self.name
        )

    def add_entity(self, e_id: str, e_attr: Dict, overwrite: bool = False):
        """Add an entity to the knowledge graph.

        Parameters
        ----------
        e_id : str
            Id of the entity you want to add.
        e_attr : Dict
            Attributes of the entity you want to add.
        overwrite : bool
            If true, overwrite existing

        Raises
        ------
        ValueError
            If entity id is already present.
        """
        if e_id in self.entities and not overwrite:
            raise ValueError(f"{e_id} already exists: {self.entities[e_id]}")
        self.entities[e_id] = e_attr

    def remove_entity(self, e_id: str):
        """Remove the entity with the id.

        Parameters
        ----------
        e_id : str
            Id of entity you want to remove.

        Raises
        ------
        KeyError
            If no entity with this id exists
        """
        del self.entities[e_id]
        if e_id in self.rel:
            del self.rel[e_id]
        if e_id in self._inv_rel:
            for other_id in self._inv_rel[e_id]:
                del self.rel[other_id][e_id]
                if len(self.rel[other_id]) == 0:
                    del self.rel[other_id]
            del self._inv_rel[e_id]

    def _add_inv_rel(self, target, source):
        if target not in self._inv_rel:
            self._inv_rel[target] = {source}
        else:
            if source not in self._inv_rel[target]:
                current_value = self._inv_rel[target]
                if not isinstance(current_value, set):
                    current_value = {current_value}
                current_value.add(source)
                self._inv_rel[target] = current_value

    def add_rel(self, source: str, target: str, value, overwrite: bool = False) -> bool:
        """Add relationhip with value.

        Parameters
        ----------
        source : str
            Entity id of source.
        target : str
            Entity id of target.
        value
            Value of relation, e.g. relation name.
        overwrite : bool
            If true, overwrites existing values for already present
            relationship, else appends the value to existing.

        Returns
        -------
        bool
            True if new information was added, else false.
        """
        if (
            source in self.rel
            and target in self.rel[source]
            and value == self.rel[source][target]
        ):
            return False
        elif source not in self.rel:
            self.rel[source] = {target: value}
            self._add_inv_rel(target, source)
        elif target not in self.rel[source]:
            self.rel[source][target] = value
            self._add_inv_rel(target, source)
        else:  # new value for existing rel
            if overwrite:
                self.rel[source][target] = value
            else:
                current_value = self.rel[source][target]
                if not isinstance(current_value, list):
                    current_value = [current_value]
                current_value.append(value)
                self.rel[source][target] = current_value
        return True

    def remove_rel(self, source: str, target: str, value=None):
        """Remove relationship or relationship value.

        Parameters
        ----------
        source : str
            Entity id of source.
        target : str
            Entity id of target.
        value
            If provided: remove only this specific value.

        Raises
        ------
        KeyError
            If relationship does not exist
        ValueError
            If value does not exist in relationship
        """
        if value is not None:
            current_value = self.rel[source][target]
            value_not_found_msg = (
                f"Cannot remove {value} from {source} -> {target}, because it is not"
                f" present in {current_value}"
            )
            if isinstance(current_value, set):
                if value not in current_value:
                    raise ValueError(value_not_found_msg)
                current_value.remove(value)
                if len(current_value) == 1:
                    current_value = next(iter(current_value))
                self.rel[source][target] = current_value
                return True
            else:
                if value != current_value:
                    raise ValueError(value_not_found_msg)
                # here we can simply remove the relationship
        del self.rel[source][target]
        if len(self.rel[source]) == 0:
            del self.rel[source]
        self._inv_rel[target].remove(source)
        if len(self._inv_rel[target]) == 0:
            del self._inv_rel[target]

    def sample(self, n: int, seed: Union[int, random.Random] = None) -> KG:
        """Return a sample of the knowledge graph with n entities.

        Parameters
        ----------
        n : int
            Number of entities to return.
        seed : Union[int, random.Random]
            Seed for randomness or seeded random.Random object.
            Default is None.

        Returns
        -------
        KG
            Knowledge graph with n entities.

        Examples
        --------
        >>> from forayer.knowledge_graph import KG
        >>> entities = {
                "e1": {"a1": "first entity", "a2": 123},
                "e2": {"a1": "second ent"},
                "e3": {"a2": 124},
            }
        >>> kg = KG(entities)
        >>> kg.sample(2)
        KG(entities={'e1': {'a1': 'first entity', 'a2': 123},
            'e2': {'a1': 'second ent'}},rel=None,name=None)
        """
        r_gen = random_generator(seed)
        sampled_e_ids = r_gen.sample(self.entities.keys(), n)
        return self.subgraph(sampled_e_ids)

    def __getitem__(self, key: Union[str, List[str]]) -> Dict[Any, Any]:
        """Return entity/entities with key(s).

        For a single key returns attributes of entity and is basically syntactic
        sugar for self.entities[key].
        For multiple keys, return a sub-dict of entities with the given ids.

        Parameters
        ----------
        key: Union[str, List[str]]
            entity id(s)

        Returns
        -------
        Dict[Any,Dict[Any,Any]]
            attributes of entity
        """
        if isinstance(key, list):
            return {e_id: self.entities[e_id] for e_id in key}
        return self.entities[key]

    def __contains__(self, key):
        # some datasets have entities without attributes that
        # only show up in the relations
        if key in self.entities or key in self.rel or key in self._inv_rel:
            return True
        return False

    def __setitem__(self, key, value):
        """Not implemented."""
        raise NotImplementedError

    def __repr__(self):
        return f"KG(entities={self.entities}, rel={self.rel}, name={self.name})"

    def neighbors(
        self, entity_id: str, only_id: bool = False
    ) -> Union[Set[str], Dict[Any, Dict[Any, Any]]]:
        """Get neighbors of an entity.

        Parameters
        ----------
        entity_id: str
            The id of entity of which we want the neighbors.
        only_id: bool
            If true only ids are returned

        Returns
        -------
        neighbors: Union[Set[str], Dict[Any, Dict[Any,Any]]]
            entity dict of neighbors, if only_id is true returns neighbor ids as set
        """
        try:
            n_to_right = set(self.rel[entity_id].keys())
        except KeyError:
            n_to_right = set()
        try:
            n_to_left = self._inv_rel[entity_id]
        except KeyError:
            n_to_left = set()
        result_ids = n_to_right.union(n_to_left)
        if only_id:
            return result_ids
        return self[list(result_ids)]

    @property
    def entity_ids(self) -> Set[str]:
        """Return ids of all entities.

        Returns
        -------
        Set[str]
            Ids of all entities.
        """
        return (
            set(self.entities.keys())
            .union(set(self.rel.keys()))
            .union(set(self._inv_rel.keys()))
        )

    @property
    def attribute_names(self) -> Set[str]:
        """Return all attribute names.

        Returns
        -------
        Set[str]
            Attribute names as set.
        """
        # get list of sets of attribute dict keys
        # and flatten into one set using chain
        return set(chain(*[set(k.keys()) for k in self.entities.values()]))

    @property
    def relation_names(self) -> Set[str]:
        """Return all relation names.

        Returns
        -------
        Set[str]
            Relation names as set.
        """
        rel_names = set()
        for _, target_rel_dict in self.rel.items():
            for _, rel_dict in target_rel_dict.items():
                if isinstance(rel_dict, str):
                    rel_names.add(rel_dict)
                elif isinstance(rel_dict, dict):
                    probably_name = list(rel_dict.keys())
                    if len(probably_name) == 1:
                        rel_names.add(probably_name[0])
        return rel_names

    def _rel_signatures(self):
        all_rels = set()
        for left, right_dict in self.rel.items():
            for right, rel_values in right_dict.items():
                if isinstance(rel_values, list):
                    for rels in rel_values:
                        if isinstance(rels, dict):
                            for inner_rel_names in rels.keys():
                                all_rels.add(left + right + str(inner_rel_names))
                        else:
                            all_rels.add(left + right + str(rels))
                elif isinstance(rel_values, dict):
                    for inner_rel_names in rel_values.keys():
                        all_rels.add(left + right + str(inner_rel_names))
                else:
                    all_rels.add(left + right + str(rel_values))
        return all_rels

    def info(self) -> str:
        """Print general information about this object.

        Returns
        -------
        str
            information about number of entities, attributes and values
        """
        num_ent = len(self.entities.keys())
        num_attr_name = len(self.entities.values())
        num_attr_values = len(
            {
                attr_value
                for _, attr_dict in self.entities.items()
                for _, attr_value in attr_dict.items()
            }
        )
        num_ent_rel = len(set(self.rel.keys()).union(set(self._inv_rel.keys())))
        num_rel = len(self._rel_signatures())
        name = "KG" if self.name is None else self.name
        return (
            f"{name}: (# entities: {len(self)}, # entities_with_rel: {num_ent_rel}, #"
            f" rel: {num_rel}, # entities_with_attributes: {num_ent}, #"
            f" attributes: {num_attr_name}, # attr_values: {num_attr_values})"
        )

    def to_rdflib(self, prefix: str = "", attr_mapping: dict = None):
        """Transform to rdflib graph.

        Parameters
        ----------
        prefix : str
            Prefix to prepend to each entity id
        attr_mapping : dict
            Mapping of attribute names to URIs.
            Mapping values can be str or :class:`rdflib.term.URIRef`.
            This is also used to map relation predicates.

        Returns
        -------
        rdf_g
            rdflib Graph

        Examples
        --------
        >>> entities = {
            "e1": {"a1": "first entity", "a2": 123},
            "e2": {"a1": "second ent"},
            "e3": {"a2": {124, "1223"}},
        }
        >>> kg = KG(entities, {"e1": {"e3": "somerelation"}})
        >>> rdf_g = kg.to_rdflib()
        >>> from rdflib import URIRef
        >>> rdf_g.value(URIRef("e1"), URIRef("a1"))
        rdflib.term.Literal('first entity')

        You can use custom prefixes and rdflib namespaces or strings for mappings

        >>> from rdflib.namespace import FOAF
        >>> my_prefix = "http://example.org/"
        >>> my_mapping = {"a1":FOAF.name, "a2":"http://example.org/attr"}
        >>> rdf_g = kg.to_rdflib(prefix=my_prefix,attr_mapping=my_mapping)
        >>> rdf_g.value(URIRef(my_prefix + "e1"), FOAF.name)
        rdflib.term.Literal('first entity')
        """

        def get_predicate(raw, attr_mapping):
            substitute = attr_mapping.get(raw, raw)
            if isinstance(substitute, URIRef):
                return substitute
            return URIRef(substitute)

        rdf_g = Graph()
        if attr_mapping is None:
            attr_mapping = {}
        for e_id, attr_dict in tqdm(
            self.entities.items(), desc="Transforming entities"
        ):
            for attr_name, attr_value in attr_dict.items():
                subject = URIRef(prefix + e_id)
                predicate = get_predicate(attr_name, attr_mapping)
                if isinstance(attr_value, (set, list)):
                    for inner_attr_val in attr_value:
                        object = Literal(inner_attr_val)
                        rdf_g.add((subject, predicate, object))
                else:
                    object = Literal(attr_value)
                    rdf_g.add((subject, predicate, object))
        for left_id, right_id_rel in self.rel.items():
            for right_id, rel in right_id_rel.items():
                subject = URIRef(prefix + left_id)
                predicate = get_predicate(rel, attr_mapping)
                object = URIRef(prefix + right_id)
                rdf_g.add((subject, predicate, object))
        return rdf_g

    def __add__(self, other):
        merged_entities = dict_merge(self.entities, other.entities)
        merged_rel = dict_merge(self.rel, other.rel)
        return KG(entities=merged_entities, rel=merged_rel)

    def __len__(self):
        # some datasets have entities without attributes that
        # only show up in the relations
        return len(
            set(self.entities)
            .union(set(self.rel.keys()))
            .union(set(self._inv_rel.keys()))
        )


class AttributeEmbeddedKG(KG):
    """KG class holding entities and their embedded attributes as well as relations between entities."""

    def __init__(
        self,
        entities: Dict[Any, Dict[Any, Any]],
        rel: Dict[Any, Dict[Any, Any]],
        vectorizer: AttributeVectorizer,
        name: str = None,
    ):
        """Initialize an attribute embeded KG object.

        Calculates the attribute embeddings given a tokenizer and vectorizer.

        Parameters
        ----------
        entities : Dict[Any, Dict[Any, Any]]
            entity information with entity ids as keys and a attribute dictionaries as values
            attribute dictionaries have attribute id as key and attribute value as dict value
        rel : Dict[Any, Dict[Any, Any]]
            relation triples with one entity as key, value is dict with other entity as key
            and relation id as value
        vectorizer: AttributeVectorizer
            an attribute vectorizer to use for retrieving the embeddings
        name : str, optional
            name of the kg, default is None

        Examples
        --------
        >>> from forayer.knowledge_graph import AttributeEmbeddedKG
        >>> from forayer.datasets import OpenEADataset
        >>> dw15kv1 = OpenEADataset(ds_pair="D_W",size="15K",version=1)
        >>> from forayer.transformation.word_embedding import AttributeVectorizer

        For demonstration we take a sample

        >>> dbpedia = dw15kv1.er_task.kgs["DBpedia"].sample(1000)

        Initialize the Vectorizer with the pre-trained embeddings

        >>> vectorizer = AttributeVectorizer(embedding_type="fasttext")

        If you have them downloaded already you can also supply the path

        >>> vectorizer = AttributeVectorizer(embedding_type="fasttext",
                vectors_path=f"somepath/fasttext/wiki.simple.bin")

        Then create a knowledge graphs with embedded attribute tokens

        >>> dbp_embedded = AttributeEmbeddedKG.from_kg(dbpedia, vectorizer=vectorizer)
        """
        self.vectorizer = vectorizer
        self.vectorizer.reset_token_count()
        attr_embedded_entities = {
            e_id: self.vectorizer.vectorize_entity_attributes(ent_attr)
            for e_id, ent_attr in tqdm(entities.items(), desc="Vectorizing attributes")
        }
        if self.vectorizer.ignored_tokens > 0:
            warnings.warn(
                f"{self.vectorizer.ignored_tokens}/{self.vectorizer.seen_tokens} tokens"
                " have no pre-trained embedding and were replaced by np.NaN"
            )
        self._ignored = self.vectorizer.ignored_tokens
        self._seen = self.vectorizer.seen_tokens
        super(AttributeEmbeddedKG, self).__init__(attr_embedded_entities, rel, name)

    def __repr__(self):
        return self.info()

    def info(self) -> str:
        """Print general information about this object.

        Returns
        -------
        str
            information about number of entities, attributes and embedded attributes
        """
        num_ent = len(self.entities.keys())
        num_attr_name = len(self.entities.values())
        num_ent_rel = len(set(self.rel.keys()).union(set(self._inv_rel.keys())))
        name = "KG" if self.name is None else self.name
        return (
            f"{name}: (# entities_with_rel: {num_ent_rel}, # rel: {len(self.rel)}, #"
            f" entities_with_attributes: {num_ent}, # attributes: {num_attr_name},"
            f" {self._ignored}/{self._seen} tokens"
            " have no pre-trained embedding and were replaced by np.NaN)"
        )

    @classmethod
    def from_kg(cls, kg: KG, vectorizer: AttributeVectorizer) -> AttributeEmbeddedKG:
        """Initialize an attribute embedded KG object from a knowledge graph object.

        Parameters
        ----------
        kg : KG
            a pre-populated knowledge graph
        vectorizer: AttributeVectorizer
            an attribute vectorizer to use for retrieving the embeddings

        Returns
        -------
        AttributeEmbeddedKG
            the given KG with vectorized attribute values
        """
        return AttributeEmbeddedKG(
            entities=kg.entities, rel=kg.rel, vectorizer=vectorizer, name=kg.name
        )



from collections import Counter, defaultdict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from forayer.knowledge_graph import KG, AttributeEmbeddedKG, ERTask
from joblib import Memory
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

memory = Memory("./data/data_frames", mmap_mode="r+")


def non_nan_embeddings(
    kg,
    *,
    percent: bool = True,
    plot: bool = True,
    sort: str = "not nan",
    sort_ascending: bool = False
):
    """Return distribution of found embeddings vs not-found embeddings per attribute.

    For each attribute counts the number of np.NaN  and non-NaN tokens in attribute values.

    Parameters
    ----------
    kg : AttributeEmbeddedKG or dict with entities
        Knowledge graph (or only entities) with attribute values as embeddings.
    percent : bool
        If true give result in percent, else absolute occurence.
    plot : bool
        If true return plot, else return pd.DataFrame.
    sort : str
       Either sort by "nan" or "not nan" if None do not sort.
    sort_ascending : bool
       If True sort ascending.

    Returns
    -------
    non_nan_embeddings
        If plot=True returns a plotly plot
        else returns a df with the information.
    """
    if isinstance(kg, AttributeEmbeddedKG):
        entities = kg.entities
    else:
        entities = kg
    nan_col = "nan"
    not_nan_col = "not nan"
    tmp_total_col = "total"
    test = defaultdict(lambda: defaultdict(int))
    for attr_dict in entities.values():
        for attr_name, attr_value in attr_dict.items():
            if len(attr_value) == 0:
                test[attr_name][nan_col] += 1
                test[attr_name][not_nan_col] = 0
            for token in attr_value:
                if isinstance(token, np.ndarray):
                    test[attr_name][not_nan_col] += 1
                else:
                    test[attr_name][nan_col] += 1
    attribute_stat_df = pd.DataFrame(test).fillna(0).T
    if percent:
        attribute_stat_df[tmp_total_col] = (
            attribute_stat_df[nan_col] + attribute_stat_df[not_nan_col]
        )
        attribute_stat_df[nan_col] = (
            attribute_stat_df[nan_col] / attribute_stat_df[tmp_total_col]
        ) * 100
        attribute_stat_df[not_nan_col] = (
            attribute_stat_df[not_nan_col] / attribute_stat_df[tmp_total_col]
        ) * 100
        del attribute_stat_df[tmp_total_col]
    if sort is not None:
        sort = sort.lower()
        attribute_stat_df = attribute_stat_df.sort_values(
            sort, ascending=sort_ascending
        )
    if plot:
        pd.options.plotting.backend = "plotly"
        return attribute_stat_df.plot(kind="bar")
    return attribute_stat_df


def attribute_distribution(
    kg, *, percent: bool = True, plot: bool = True, sort_ascending: bool = False
):
    """Calculate the occurence of attribute names across entities.

    Parameters
    ----------
    kg : KG or dict with entities.
        Knowledge graph (or only entities) with attribute values as embeddings.
    percent : bool
        If true give result in percent of all attributes, else absolute occurence.
    plot : bool
        If true return plot, else return pd.DataFrame
    sort_ascending : bool
        If true sort_ascending on occurence else, descending

    Returns
    -------
    attr_distribution
        If plot=True returns a plotly plot
        else returns a df with the information.
    """
    if isinstance(kg, KG):
        entities = kg.entities
    else:
        entities = kg
    col_name = "occurence"
    attribute_count = Counter(k for attr in entities.values() for k in attr.keys())
    (index, values) = zip(*dict(attribute_count).items())
    attribute_stat_df = pd.DataFrame(values, columns=[col_name], index=index).fillna(0)
    if percent:
        attribute_stat_df = (attribute_stat_df / len(entities)) * 100
        new_col_name = "present in % of entities"
        attribute_stat_df.rename(columns={col_name: new_col_name}, inplace=True)
        col_name = new_col_name
    attribute_stat_df = attribute_stat_df.sort_values(
        col_name, ascending=sort_ascending
    )
    if plot:
        pd.options.plotting.backend = "plotly"
        return attribute_stat_df.plot(kind="bar")
    return attribute_stat_df


@memory.cache
def _reduce_dim(embeddings):
    return TSNE(n_components=2, n_jobs=-1).fit_transform(
        PCA(n_components=20).fit_transform(embeddings)
    )


def _reduced_embeddings(entities):
    wanted_attr = set(non_nan_embeddings(entities, percent=False, plot=False).index)
    attr_indices = defaultdict(list)
    emb_list = []
    attr_name_to_ent = {}
    for e_id, e_attr_dict in entities.items():
        for a_name, e_attr_value in e_attr_dict.items():
            attr_name_to_ent[a_name] = e_id
            if a_name in wanted_attr and len(wanted_attr) > 0:
                for token_emb in e_attr_value:
                    if not isinstance(token_emb, float):
                        emb_list.append(token_emb)
                        attr_indices[a_name].append(len(emb_list) - 1)
    attr_indices = dict(attr_indices)
    embeddings = np.array(emb_list)
    return _reduce_dim(embeddings), attr_indices, attr_name_to_ent


def _er_task_embedding_df_creation(emb2d, attr_indices, attr_name_to_ent, er_task):
    row = []
    for attr_name, indices in attr_indices.items():
        for _ in indices:
            ent_id = attr_name_to_ent[attr_name]
            # lone entities are assigne to cluster_id -1
            cluster_id = er_task.clusters.get(ent_id, value=-1)
            row.append((ent_id, cluster_id, attr_name))
    attr_emb_dim_red = pd.DataFrame(
        row, columns=["entity_id", "cluster_id", "attribute"]
    )
    emb2d_df = pd.DataFrame(emb2d, columns=["embx", "emby"])
    return attr_emb_dim_red.join(emb2d_df), "cluster_id"


def _embedding_df_creation(emb2d, attr_indices, attr_name_to_ent, entities):
    row = []
    for attr_name, indices in attr_indices.items():
        for _ in indices:
            row.append((attr_name_to_ent[attr_name], attr_name))
    attr_emb_dim_red = pd.DataFrame(row, columns=["entity_id", "attribute"])
    emb2d_df = pd.DataFrame(emb2d, columns=["embx", "emby"])
    return attr_emb_dim_red.join(emb2d_df), "entity_id"


def dim_reduced_embeddings_df(entities):
    """Create a DataFrame with x,y coordinates of dimensionality reduced attribute embeddings.

    Parameters
    ----------
    entities
        Entities with attribute embeddings.

    Returns
    -------
    pd.DataFrame
        df with the columns 'attribute,embx,emby'
    color_col_suggestion
        suggested column for color
    """
    df_creation = _embedding_df_creation
    if isinstance(entities, KG):
        _entities = entities.entities
    elif isinstance(entities, ERTask):
        _entities = entities.all_entities()
        df_creation = _er_task_embedding_df_creation

    emb2d, attr_indices, attr_name_to_ent = _reduced_embeddings(_entities)
    return df_creation(emb2d, attr_indices, attr_name_to_ent, entities)


def plot_attribute_embeddings(entities, subset=None, color_col=None, colorscale=None):
    """Plot attribute embeddings.

    Parameters
    ----------
    entities
        Entities with embedded attributes to plot.
        Can be KG or ERTask.
    subset
        Specify a subset of entity ids to plot, but create the
        dimensionality redcution based on all provided.
    color_col
        Column to use for color. Can be "entity_id, cluster_id or attribute"
    colorscale
        Use a specific color scale
    """
    attr_emb_dim_red, color_col_suggestion = dim_reduced_embeddings_df(entities)
    if subset is not None:
        attr_emb_dim_red = attr_emb_dim_red[attr_emb_dim_red["entity_id"].isin(subset)]
    if color_col is None and color_col_suggestion == "cluster_id":
        actual_color_col = "cluster_id"
    else:
        if color_col == "is_clustered":
            numeric_color = {
                attr: i
                for i, attr in enumerate(list(attr_emb_dim_red["cluster_id"].unique()))
            }
            attr_emb_dim_red["color"] = attr_emb_dim_red[color_col].replace(
                numeric_color
            )
            actual_color_col = "color"
        else:
            color_col = color_col_suggestion if color_col is None else color_col
            numeric_color = {
                attr: i
                for i, attr in enumerate(list(attr_emb_dim_red[color_col].unique()))
            }
            attr_emb_dim_red["color"] = attr_emb_dim_red[color_col].replace(
                numeric_color
            )
            actual_color_col = "color"

    return go.Figure(
        data=go.Scatter(
            x=attr_emb_dim_red["embx"],
            y=attr_emb_dim_red["emby"],
            mode="markers",
            marker_color=attr_emb_dim_red[actual_color_col],
            text=attr_emb_dim_red["attribute"],
            marker_colorscale=colorscale,
        )
    )



"""Help with randomn functions."""
import random
from typing import Union


def random_generator(seed: Union[int, random.Random] = None) -> random.Random:
    """Create or return random generator.

    Parameters
    ----------
    seed : Union[int, random.Random]
        Either a random generator or a seed or None.
    Returns
    -------
    random.Random
        If None was provided return a randomly seeded
        random generator. If a seed was provideded return
        a generator with this seed. If a random.Random object
        was provided return it.
    """
    if seed is not None:
        if isinstance(seed, int):
            r_gen = random.Random(seed)
        elif isinstance(seed, random.Random):
            r_gen = seed
    else:
        r_gen = random.Random(seed)
    return r_gen



"""Read and write semantic web sources."""
import logging
from collections import defaultdict
from typing import Callable, Set

import rdflib
from forayer.knowledge_graph import KG
from rdflib import Graph
from rdflib.term import Literal
from tqdm import tqdm


def load_from_rdf(
    in_path: str,
    literal_cleaning_func: Callable = None,
    format: str = None,
    kg_name: str = None,
    multi_value: Callable = None,
) -> KG:
    """Create knowledge graph object from rdf source.

    Parameters
    ----------
    in_path : str
        Path of triple file.
    literal_cleaning_func: Callable
        Function to preprocess literals,
        if None will simply cast to python types.
    format : str
        Triple format ("xml”, “n3” (use for turtle), “nt” or “trix”).
    kg_name : str
        How to name the knowledge graph object.
    multi_value : Callable
        How to handle multiple attribute values for an
        entity, attribute name combination.
        Default creates a set and adds to it

    Returns
    -------
    KG
        the loaded kg object
    """
    if literal_cleaning_func is None:
        literal_cleaning_func = cast_to_python_type
    if multi_value is None:
        multi_value = add_multi_value
    g = Graph()
    logging.info(f"Reading graph from {in_path}. This might take a while...")
    g.parse(in_path, format=format)
    entities = defaultdict(dict)
    rel = defaultdict(dict)
    for stmt in tqdm(g, desc="Transforming graph", total=len(g)):
        s, p, o = stmt
        if isinstance(o, Literal):
            value = literal_cleaning_func(o)
            if str(p) in entities[str(s)]:
                value = multi_value(entities[str(s)][str(p)], value)
            entities[str(s)][str(p)] = value
        else:
            rel[str(s)][str(o)] = str(p)
    return KG(entities=entities, rel=rel, name=kg_name)


def write_to_rdf(
    kg: KG, out_path: str, format: str, prefix: str = "", attr_mapping: dict = None
):
    """Write the forayer knowledge graph to a rdf serialization format.

    Parameters
    ----------
    kg : KG
        The knowledge graph that will be serialized.
    out_path : str
        The path where it should be serialized to.
    format : str
        The desired rdf format.
    prefix : str
        Prefix for the entities in the graph.
    attr_mapping : dict
       Mapping of attribute names.
    """
    kg.to_rdflib(prefix=prefix, attr_mapping=attr_mapping).serialize(
        destination=out_path, format=format
    )


def cast_to_python_type(lit: rdflib.term.Literal):
    """Casts a literal to the respective python type.

    Parameters
    ----------
    lit : rdflib.term.Literal
       The literal that is to be cast.

    Returns
    -------
    Any
        The literal as the respective python object
    """
    if lit.datatype is not None and "langString" in lit.datatype:
        # lang strings are not automatically cast
        return str(lit)
    return lit.toPython()


def add_multi_value(prev, new) -> Set:
    """Add a  value to a set or create a new set with prev and new.

    Parameters
    ----------
    prev
        Existing value.
    new
        New value, that should be added.

    Returns
    -------
    Set
        Set containing the previous and new elements.
    """
    if not isinstance(prev, set):
        prev = {prev}
    prev.add(new)
    return prev



"""Helper for more complicated dict functions."""
from collections import defaultdict


# see https://stackoverflow.com/a/30921635
def nested_ddict():
    """Create a nested defaultdict."""
    return defaultdict(nested_ddict)


# see https://stackoverflow.com/a/30921635
def nested_ddict2dict(d):
    """Cast a nested defaultdict to dict."""
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = nested_ddict2dict(v)
    return dict(d)


def _merge(a, b, path=None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if (
                not (isinstance(a[key], dict) and isinstance(b[key], dict))
                or a[key] == b[key]
            ):
                raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
            elif isinstance(a[key], dict) and isinstance(b[key], dict):
                dict_merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
        else:
            a[key] = b[key]
    return a


# see https://stackoverflow.com/a/7205107
def dict_merge(a, b):
    """Merge a and b.

    Parameters
    ----------
    a
        One dictionary that will be merged
    b
        Other dictionary that will be merged
    """
    return _merge(dict(a), b)



"""OpenEA dataset class."""
import os

from forayer.datasets.base_dataset import ZipDataset
from forayer.input_output.from_to_open_ea import from_openea
from forayer.knowledge_graph import ERTask


class OpenEADataset(ZipDataset):
    """The OpenEA datasets contain entity resolution tasks with samples from popular knowledge graphs.

    Several different tasks are available with snippets from DBpedia, Wikidata and YAGO.
    Different sizes refer to the number of entities in the respective graphs (15K or 100K).
    For each setting two versions are available, where version 1 has lower connectivity
    in the graph compared to version 2.

    More information can be found at the respective `github repository <https://github.com/nju-websoft/OpenEA>`_ and
    the benchmark publication: Sun et al (2020) `A Benchmarking Study of Embedding-based Entity Alignment for Knowledge Graphs`, *VLDB* <http://www.vldb.org/pvldb/vol13/p2326-sun.pdf>
    """

    DS_NAME = "OpenEA"

    __DOWNLOAD_URLS = [
        (
            "https://www.dropbox.com/s/xfehqm4pcd9yw0v/OpenEA_dataset_v2.0.zip?dl=1",
            "OpenEA_dataset_v2.0.zip",
        )
    ]

    def __init__(
        self,
        ds_pair: str,
        size: str,
        version: int,
        data_folder: str = None,
    ):
        """Initialize an OpenEA dataset pair.

        Parameters
        ----------
        ds_pair : str
            name of ds pair (either "D-W" or "D-Y")
        size : str
            size of the task (either "15K" or "100K")
        version : int
            version of task (either 1 or 2)
        data_folder : str
            folder where raw files are stored or will be downloaded
        """
        super(OpenEADataset, self).__init__(
            data_folder=data_folder, download_urls=self.__class__.__DOWNLOAD_URLS
        )
        self.ds_pair = ds_pair
        self.size = size
        self.version = version
        self.download_and_unzip()
        self.er_task = self._load()

    def __repr__(self):
        return (
            self.__class__.__name__
            + f"(ds_pair={self.ds_pair}, size={self.size}, version={self.version},"
            f" data_folder={self.data_folder}, {self.er_task})"
        )

    def _load(self) -> ERTask:
        """Load :class:`ERTask` object from raw files.

        Returns
        -------
        ERTask
            The er task created from the files
        """
        task_folder = os.path.join(
            self.data_folder,
            self.__class__.__DOWNLOAD_URLS[0][1].replace(".zip", ""),
            f"{self.ds_pair}_{self.size}_V{self.version}",
        )
        return from_openea(task_folder)



"""Write Gradoop file format and read Gradoop file format."""
import csv
import os
from collections import defaultdict, namedtuple
from typing import Dict, List, Tuple, Union

from forayer.knowledge_graph import KG
from forayer.utils.dict_help import nested_ddict, nested_ddict2dict

TYPE_CONVERSION = {
    "string": str,
    "long": int,
    "int": int,
    "float": float,
    "double": float,
    "boolean": lambda x: x == "true",
}

INV_TYPES = {
    str: "string",
    int: "int",
    float: "float",
    bool: "boolean",
}

VertexLine = namedtuple("VertexLine", ["id", "graph_ids", "type", "props"])
EdgeLine = namedtuple(
    "EdgeLine", ["id", "graph_ids", "source_id", "target_id", "type", "props"]
)


def int_to_gradoop_id(value: int) -> str:
    """Casts to Gradoop id, which are 12 byte hexadecimal strings.

    Parameters
    ----------
    value : int
        Value to cast

    Returns
    -------
    str
        12 byte hexadecimal string (without leading '0x').
    """
    # see https://stackoverflow.com/a/12638477
    # first two characters are leading 0x
    return f"{value:#0{26}x}"[2:]


def is_gradoop_id(value) -> bool:
    """Check if value is a valid Gradoop id.

    Gradoop ids are 12 byte hexadecimal strings

    Parameters
    ----------
    value
        Value to check

    Returns
    -------
    bool
        True if is valid Gradoop id
    """
    if isinstance(value, str) and len(value) == 24:
        try:
            int(value, 16)
            return True
        except ValueError:
            return False
    return False


def _load_metadata(path: str) -> Dict[str, Dict[str, List[Tuple[str, str]]]]:
    metadata = {"g": defaultdict(list), "v": defaultdict(list), "e": defaultdict(list)}

    with open(path, "r") as in_file:
        reader = csv.reader(in_file, delimiter=";")
        for row in reader:
            if len(row) == 2:
                gve, label = row
                properties = []
            else:
                gve, label, keys = row
                properties = [tuple(p.split(":")) for p in keys.split(",")]
            if gve not in ["g", "v", "e"]:
                raise ValueError(
                    f"Unknown node type specifier '{gve}', only 'g','v' or 'e' allowed"
                )
            metadata[gve][label] = properties
    # cast to normal dict
    return {
        "g": dict(metadata["g"]),
        "v": dict(metadata["v"]),
        "e": dict(metadata["e"]),
    }


def _prop_creation(element_type, label, metadata, properties, is_edge=False):
    props = {} if is_edge else {"_label": label}
    for prop_name_type, prop_value in zip(
        metadata[element_type][label], properties.split("|")
    ):
        if len(prop_name_type) == 1:
            continue
        prop_name, prop_type = prop_name_type
        if prop_value != "":
            props[prop_name] = TYPE_CONVERSION[prop_type](prop_value)
    return props


def _load_graphs(path: str, metadata: Dict[str, Dict[str, List[Tuple[str, str]]]]):
    graphs = {}
    with open(path, "r") as in_file:
        reader = csv.reader(in_file, delimiter=";")
        for row in reader:
            g_id, label, properties = row
            graphs[g_id] = _prop_creation("g", label, metadata, properties)
    return graphs


def _graph_containment(graphs: str):
    return graphs[1:-1].split(",")


def _load_vertices(path: str, metadata: Dict[str, Dict[str, List[Tuple[str, str]]]]):
    graph_vertices = nested_ddict()
    with open(path, "r") as in_file:
        reader = csv.reader(in_file, delimiter=";")
        for row in reader:
            v_id, graphs, label, properties = row
            v_props = _prop_creation("v", label, metadata, properties)
            for g in _graph_containment(graphs):
                graph_vertices[g][v_id] = v_props
    return nested_ddict2dict(graph_vertices)


def _load_edges(path: str, metadata: Dict[str, Dict[str, List[Tuple[str, str]]]]):
    graph_edges = nested_ddict()
    with open(path, "r") as in_file:
        reader = csv.reader(in_file, delimiter=";")
        for row in reader:
            e_id, graphs, source, target, label, properties = row
            e_props = _prop_creation("e", label, metadata, properties, is_edge=True)
            for g in _graph_containment(graphs):
                graph_edges[g][source][target][label] = e_props
    return nested_ddict2dict(graph_edges)


def load_from_csv_datasource(
    folder_path: str, graph_name_property: str = None
) -> Dict[str, KG]:
    """Load Gradoop graph from csv datasource.

    Parameters
    ----------
    folder_path : str
        Path for folder that contains graph.
    graph_name_property : str
        Name of graph property that will be used to name graphs.
        If None use graph id.

    Returns
    -------
    Dict[str,KG]
        Dictionary of knowledge graphs.
    """
    graphs_csv_path = os.path.join(folder_path, "graphs.csv")
    vertices_csv_path = os.path.join(folder_path, "vertices.csv")
    edges_csv_path = os.path.join(folder_path, "edges.csv")
    metadata_csv_path = os.path.join(folder_path, "metadata.csv")

    metadata = _load_metadata(metadata_csv_path)
    graphs = _load_graphs(graphs_csv_path, metadata)
    vertices = _load_vertices(vertices_csv_path, metadata)
    edges = _load_edges(edges_csv_path, metadata)
    if graph_name_property is not None:
        return {
            g: KG(
                entities=vertices[g],
                rel=edges[g],
                name=graphs[g][graph_name_property],
            )
            for g in graphs.keys()
        }
    else:
        return {
            g: KG(entities=vertices[g], rel=edges[g], name=g) for g in graphs.keys()
        }


def _gather_edge_metadata(kgs: Dict[str, KG], attribute_type_mapping: Dict = None):
    e_metadata = nested_ddict()
    for kg in kgs.values():
        for _, rel_dict in kg.rel.items():
            for _, rel_prop_dict in rel_dict.items():
                if isinstance(rel_prop_dict, str):
                    # in this case
                    # relation does not have attributes
                    # and simply has the relation name
                    # which we use as edge type for gradoop
                    e_metadata[rel_prop_dict] = {}
                    continue
                for e_type, inner_prop_dict in rel_prop_dict.items():
                    for e_prop_name, e_prop_value in inner_prop_dict.items():
                        if (
                            attribute_type_mapping is not None
                            and e_type in attribute_type_mapping
                            and e_prop_name in attribute_type_mapping[e_type]
                        ):
                            e_metadata[e_type][e_prop_name] = attribute_type_mapping[
                                e_type
                            ][e_prop_name]
                        elif e_metadata[e_type][e_prop_name] != {} and e_metadata[
                            e_type
                        ][e_prop_name] != type(e_prop_value):
                            raise ValueError(
                                f"Inconsistent typing for {e_prop_name} in relation"
                                f" type {e_type}"
                            )
                        else:
                            e_metadata[e_type][e_prop_name] = type(e_prop_value)
    return nested_ddict2dict(e_metadata)


def _gather_vertex_metadata(
    kgs: Dict[str, KG],
    label_attr: str = "_label",
    attribute_type_mapping: Dict = None,
    vertex_id_attr_name: str = None,
):
    v_metadata = nested_ddict()
    for kg in kgs.values():
        for e_id, e_attr_dict in kg.entities.items():
            if label_attr not in e_attr_dict:
                raise ValueError(
                    f"Entity {e_id} does not contain the required label attribute"
                    f" '{label_attr}'"
                )
            else:
                cur_label = e_attr_dict[label_attr]
            for attr_name, attr_val in e_attr_dict.items():
                if attr_name == label_attr:
                    continue
                elif (
                    attribute_type_mapping is not None
                    and cur_label in attribute_type_mapping
                    and attr_name in attribute_type_mapping[cur_label]
                ):
                    v_metadata[cur_label][attr_name] = attribute_type_mapping[
                        cur_label
                    ][attr_name]
                elif v_metadata[cur_label][attr_name] != {} and v_metadata[cur_label][
                    attr_name
                ] != type(attr_val):
                    raise ValueError(
                        f"Inconsistent typing for {attr_name} in Entity type"
                        f" {cur_label}"
                    )
                else:
                    v_metadata[cur_label][attr_name] = type(attr_val)
                if vertex_id_attr_name is not None:
                    v_metadata[cur_label][vertex_id_attr_name] = str
    return nested_ddict2dict(v_metadata)


def _fix_metadata_order(metadata: Dict):
    fixed_metadata = {}
    for element_type, ele_meta in metadata.items():
        v_out = {}
        for inner_type, attr_dict in ele_meta.items():
            inner_list = [[], []]  # name, type
            for a_name, a_type in attr_dict.items():
                inner_list[0].append(a_name)
                inner_list[1].append(a_type)
            v_out[inner_type] = inner_list
        fixed_metadata[element_type] = v_out
    return fixed_metadata


def _create_metadata(
    kgs: Dict[str, KG],
    label_attr: str = "_label",
    attribute_type_mapping: Dict = None,
    vertex_id_attr_name: str = None,
    graph_name_as_property: str = None,
):
    edge_type_mapping = None
    vertex_type_mapping = None
    if attribute_type_mapping is not None:
        edge_type_mapping = (
            None if "e" not in attribute_type_mapping else attribute_type_mapping["e"]
        )
        vertex_type_mapping = (
            None if "v" not in attribute_type_mapping else attribute_type_mapping["v"]
        )
    edge_metadata = _gather_edge_metadata(kgs, attribute_type_mapping=edge_type_mapping)

    vertex_metadata = _gather_vertex_metadata(
        kgs,
        label_attr=label_attr,
        attribute_type_mapping=vertex_type_mapping,
        vertex_id_attr_name=vertex_id_attr_name,
    )

    graph_metadata = {}
    for i, k in enumerate(kgs.values(), start=1):
        props = {} if graph_name_as_property is None else {graph_name_as_property: str}
        if k.name is None:
            graph_metadata[f"graph{i}"] = props
        else:
            graph_metadata[k.name] = props

    return _fix_metadata_order(
        {"g": graph_metadata, "e": edge_metadata, "v": vertex_metadata}
    )


def _create_vertex_lines(
    kgs: Dict[str, KG], label_attr: str, vertex_metadata: Dict, vertex_id_attr_name: str
):
    v_dict = {}
    vid_to_gid = {}
    for k_name, kg in kgs.items():
        for e_id, e_attr_dict in kg.entities.items():
            cur_label = e_attr_dict[label_attr]
            prop_line = []
            for attr_name, exp_type in zip(*vertex_metadata[cur_label]):
                if attr_name == vertex_id_attr_name:
                    attr_value = e_id
                else:
                    attr_value = e_attr_dict.get(attr_name, "")
                if attr_value != "":
                    if exp_type == bool:
                        attr_value = "true" if attr_value else "false"
                    elif not isinstance(exp_type, str):  # custom type
                        attr_value = exp_type(attr_value)
                prop_line.append(str(attr_value))
            prop_string = "|".join(prop_line)
            if e_id in v_dict:
                if v_dict[e_id].props != prop_string:
                    raise ValueError(
                        f"Entity {e_id} has inconsistent representation across"
                        f" graphs:{prop_string}\n and\n {v_dict[e_id][3]}"
                    )
                else:
                    v_dict[e_id].graph_ids.append(k_name)
            else:
                if not is_gradoop_id(e_id):
                    grad_id = int_to_gradoop_id(len(v_dict))
                    vid_to_gid[e_id] = grad_id
                else:
                    grad_id = e_id
                v_dict[e_id] = VertexLine(grad_id, [k_name], cur_label, prop_string)
    return list(v_dict.values()), vid_to_gid


def _create_edge_lines(kgs: Dict[str, KG], edge_metadata: Dict, vid_to_gid: Dict):
    e_dict = {}
    for k_name, kg in kgs.items():
        for source_id, target_rel_dict in kg.rel.items():
            for target_id, rel_dict in target_rel_dict.items():
                if isinstance(rel_dict, str):
                    # in this case
                    # relation does not have attributes
                    # and simply has the relation name
                    # which we use as edge type for gradoop
                    rel_dict = {rel_dict: {}}
                for cur_label, prop_dict in rel_dict.items():
                    prop_line = []
                    for attr_name, exp_type in zip(*edge_metadata[cur_label]):
                        attr_value = prop_dict.get(attr_name, "")
                        if attr_value != "":
                            if exp_type == bool:
                                attr_value = "true" if attr_value else "false"
                            elif not isinstance(exp_type, str):  # custom type
                                attr_value = exp_type(attr_value)
                        prop_line.append(str(attr_value))
                    tmp_id = str(source_id) + str(target_id) + str(cur_label)
                    prop_string = "|".join(prop_line)
                    if tmp_id in e_dict:
                        e_dict[tmp_id].graph_ids.append(k_name)
                    else:
                        edge_id = int_to_gradoop_id(len(e_dict))
                        if vid_to_gid is not None:
                            source_id = vid_to_gid.get(source_id, source_id)
                            target_id = vid_to_gid.get(target_id, target_id)
                        e_dict[tmp_id] = EdgeLine(
                            edge_id,
                            [k_name],
                            source_id,
                            target_id,
                            cur_label,
                            prop_string,
                        )
    return list(e_dict.values())


def _create_graph_lines(
    kgs: Dict[str, KG],
    graph_metadata: Dict,
    default_type="graph",
    graph_name_as_property: str = None,
):
    g_lines = []
    for g_id, kg in kgs.items():
        if kg.name is None:
            g_lines.append((g_id, default_type, ""))
        else:
            metadata = graph_metadata[kg.name]
            # TODO graphs cannot really have attributes yet
            props = []
            for m_att in metadata[0]:
                if (
                    graph_name_as_property is not None
                    and m_att == graph_name_as_property
                ):
                    props.append(kg.name)
                else:
                    props.append("")
            prop_string = "|".join(props)
            g_lines.append((g_id, kg.name, prop_string))
    return g_lines


def _create_metadata_lines(metadata):
    m_lines = []
    for ele_type, ele_dict in metadata.items():
        for inner_type, prop_list in ele_dict.items():
            props_list = []
            for prop_name, type_class in zip(*prop_list):
                if isinstance(type_class, str):
                    props_list.append(f"{prop_name}:{type_class}")
                else:
                    props_list.append(f"{prop_name}:{INV_TYPES[type_class]}")
            props = ",".join(props_list)
            m_lines.append((ele_type, inner_type, props))
    return m_lines


def _kgs_dict_to_gradoop_id(kgs: Dict) -> Dict:
    return {
        int_to_gradoop_id(i): kg_name_value[1]
        for i, kg_name_value in enumerate(kgs.items())
    }


def _write_lines(lines, out_path):
    with open(out_path, "w") as out_file:
        for line in lines:
            if isinstance(line, (VertexLine, EdgeLine)):
                line = list(line)
                line[1] = "[" + ",".join(line[1]) + "]"
            out_file.write(";".join([str(ele) for ele in line]) + "\n")


def write_to_csv_datasource(
    kgs: Union[KG, Dict[str, KG]],
    out_path: str,
    label_attr: str = "_label",
    attribute_type_mapping: Dict = None,
    vertex_id_attr_name: str = "_forayer_id",
    default_graph_type: str = "graph",
    graph_name_as_property: str = None,
    overwrite: bool = False,
):
    """Write knowledge graph(s) to Gradoop CSV Datasource.

    Parameters
    ----------
    kgs : Union[KG, Dict[str, KG]]
        Knowledge Graph(s) to serialize.
    out_path : str
        Folder where this data will be serialized to.
    label_attr : str, Default = "_label"
        Vertex attribute to use for Gradoop's special type attribute.
    attribute_type_mapping : Dict, Default=None
        Manually set attribute types.
    vertex_id_attr_name : str, Default="_forayer_id"
        Save the current entity id as property with this name.
        If set to None, entity id is not saved.
    default_graph_type : str
        Label graphs as this type if they do not have a name.
    graph_name_as_property : str
        Save the name of graphs as seperate property.
    overwrite : bool
        If True, overwrites existing files at output.
    """
    if os.path.exists(out_path) and not overwrite:
        raise ValueError(f"Path {out_path} already exists")
    os.makedirs(out_path, exist_ok=True)
    graphs_csv_path = os.path.join(out_path, "graphs.csv")
    vertices_csv_path = os.path.join(out_path, "vertices.csv")
    edges_csv_path = os.path.join(out_path, "edges.csv")
    metadata_csv_path = os.path.join(out_path, "metadata.csv")
    metadata = _create_metadata(
        kgs=kgs,
        label_attr=label_attr,
        attribute_type_mapping=attribute_type_mapping,
        vertex_id_attr_name=vertex_id_attr_name,
        graph_name_as_property=graph_name_as_property,
    )
    kg_dict_with_gid = _kgs_dict_to_gradoop_id(kgs)
    vertex_lines, vid_to_gid = _create_vertex_lines(
        kgs=kg_dict_with_gid,
        label_attr=label_attr,
        vertex_metadata=metadata["v"],
        vertex_id_attr_name=vertex_id_attr_name,
    )

    edge_lines = _create_edge_lines(
        kgs=kg_dict_with_gid, edge_metadata=metadata["e"], vid_to_gid=vid_to_gid
    )
    graph_lines = _create_graph_lines(
        kgs=kg_dict_with_gid,
        graph_metadata=metadata["g"],
        default_type=default_graph_type,
        graph_name_as_property=graph_name_as_property,
    )
    metadata_lines = _create_metadata_lines(metadata)
    _write_lines(graph_lines, graphs_csv_path)
    _write_lines(edge_lines, edges_csv_path)
    _write_lines(vertex_lines, vertices_csv_path)
    _write_lines(metadata_lines, metadata_csv_path)



"""OAEI knowledge graph track dataset class."""

import os
from collections import namedtuple

from forayer.datasets.base_dataset import RemoteDataset
from forayer.input_output.from_to_rdf import load_from_rdf
from forayer.knowledge_graph import ClusterHelper, ERTask
from rdflib import Graph
from tqdm import tqdm

URLWithFileName = namedtuple("URLWithFileName", ["url", "file_name"])
OAEITaskFiles = namedtuple(
    "OAEITaskFiles",
    [
        "kg1",
        "kg2",
        "ref",
    ],
)


class OAEIKGDataset(RemoteDataset):
    """The  OAEI (Ontology Alignment Evaluation Initiative) Knowledge Graph Track tasks contain graphs created from fandom wikis.

    Five integration tasks are available:
        - starwars-swg
        - starwars-swtor
        - marvelcinematicuniverse-marvel
        - memoryalpha-memorybeta
        - memoryalpha-stexpanded

    More information can be found at the `website <http://oaei.ontologymatching.org/2019/knowledgegraph/index.html>`_.
    """

    DS_NAME = "OAEI_KG_Track"

    _STARWARS = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/starwars-swg/component/source/",
        file_name="starwars.xml",
    )
    _TOR = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/starwars-swtor/component/target/",
        file_name="swtor.xml",
    )
    _SWG = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/starwars-swg/component/target/",
        file_name="swg.xml",
    )
    _MARVEL = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/marvelcinematicuniverse-marvel/component/target/",
        file_name="marvel.xml",
    )
    _MCU = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/marvelcinematicuniverse-marvel/component/source/",
        file_name="marvelcinematicuniverse.xml",
    )
    _MEMORY_ALPHA = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/memoryalpha-stexpanded/component/source/",
        file_name="memoryalpha.xml",
    )
    _STEXPANDED = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/memoryalpha-stexpanded/component/target/",
        file_name="stexpanded.xml",
    )
    _MEMORY_BETA = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/memoryalpha-memorybeta/component/target/",
        file_name="memorybeta.xml",
    )
    _SW_SWG = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/starwars-swg/component/reference.xml",
        file_name="ref-starwars-swg.xml",
    )
    _SW_TOR = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/starwars-swtor/component/reference.xml",
        file_name="ref-starwars-swtor.xml",
    )
    _MCU_MARVEL = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/marvelcinematicuniverse-marvel/component/reference.xml",
        file_name="ref-marvelcinematicuniverse-marvel.xml",
    )
    _MEMORY_ALPHA_BETA = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/memoryalpha-memorybeta/component/reference.xml",
        file_name="ref-memoryalpha-memorybeta.xml",
    )

    _MEMORY_ALPHA_STEXPANDED = URLWithFileName(
        url="http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph/v3/suite/memoryalpha-stexpanded/component/reference.xml",
        file_name="ref-memoryalpha-stexpanded.xml",
    )

    _TASKS = {
        "starwars-swg": OAEITaskFiles(kg1=_STARWARS, kg2=_SWG, ref=_SW_SWG),
        "starwars-swtor": OAEITaskFiles(kg1=_STARWARS, kg2=_TOR, ref=_SW_TOR),
        "marvelcinematicuniverse-marvel": OAEITaskFiles(
            kg1=_MCU, kg2=_MARVEL, ref=_MCU_MARVEL
        ),
        "memoryalpha-memorybeta": OAEITaskFiles(
            kg1=_MEMORY_ALPHA, kg2=_MEMORY_BETA, ref=_MEMORY_ALPHA_BETA
        ),
        "memoryalpha-stexpanded": OAEITaskFiles(
            kg1=_MEMORY_ALPHA, kg2=_STEXPANDED, ref=_MEMORY_ALPHA_STEXPANDED
        ),
    }

    def __init__(self, task: str, data_folder: str = None):
        """Initialize a OAEI Knowledge Graph Track task.

        Parameters
        ----------
        task : str
            Name of the task.
            Has to be one of {starwars-swg,starwars-swtor,marvelcinematicuniverse-marvel,memoryalpha-memorybeta, memoryalpha-stexpanded}
        data_folder : str
            folder where raw files are stored or will be downloaded
        """
        if not task.lower() in self.__class__._TASKS:
            raise ValueError(
                f"Task has to be one of {self.__class__._TASKS}, but got{task}"
            )
        task = task.lower()
        self.task = task
        dl_task = self.__class__._TASKS[self.task]
        super(OAEIKGDataset, self).__init__(
            download_urls=[dl_task.kg1, dl_task.kg2, dl_task.ref],
            data_folder=data_folder,
        )
        self.download()
        self.er_task = self._load()

    def __repr__(self):
        return (
            self.__class__.__name__
            + f"(task={self.task}, data_folder={self.data_folder})"
        )

    def _load_entity_links(self, path: str) -> ClusterHelper:
        print("Parsing ref graph")
        ref = Graph()
        ref.parse(path, format="xml")
        pairs = {}
        for stmt in tqdm(ref, desc="Gathering links"):
            s, p, o = stmt
            if (
                "http://knowledgeweb.semanticweb.org/heterogeneity/alignmententity"
                in str(p)
            ):
                # key is BNode id, value is entity id tuple
                tup = (None, None)
                if str(s) in pairs:
                    tup = pairs[str(s)]
                if "alignmententity1" in str(p):
                    tup = (str(o), tup[1])
                else:
                    tup = (tup[0], str(o))
                pairs[str(s)] = tup
        links = [{t[0], t[1]} for _, t in pairs.items()]
        return ClusterHelper(links)

    def _load(self) -> ERTask:
        dl_task = self.__class__._TASKS[self.task]
        left_name = dl_task.kg1.file_name.replace(".xml", "")
        right_name = dl_task.kg2.file_name.replace(".xml", "")
        kg_left = load_from_rdf(
            os.path.join(self.data_folder, dl_task.kg1.file_name),
            format="xml",
            kg_name=left_name,
        )
        kg_right = load_from_rdf(
            os.path.join(self.data_folder, dl_task.kg2.file_name),
            format="xml",
            kg_name=right_name,
        )
        ref_links = self._load_entity_links(
            os.path.join(self.data_folder, dl_task.ref.file_name)
        )
        return ERTask(kgs=[kg_left, kg_right], clusters=ref_links)



"""IO module for OpenEA data."""
import os
from collections import defaultdict
from typing import Dict

from forayer.knowledge_graph import KG, ClusterHelper, ERTask


def _get_cleaned_split(line: str, delimiter: str):
    seperated = line.split(delimiter)
    seperated[-1] = seperated[-1].strip()
    return seperated


def read_attr_triples(path: str, delimiter="\t") -> Dict[str, Dict[str, str]]:
    """Read attribute triples from csv into a dictionary.

    This functions returns the triples as dictionary, where entity ids are
    keys and the values are attribute dictionaries, with the attribute name
    as key.

    Parameters
    ----------
    path : str
        Path to the file
    delimiter: str, default = tab
        Delimiter of the csv file

    Returns
    -------
    ent_attr_dict: Dict[str, Dict[str, str]]
        Entity and attribute dictionary

    """
    ent_attr_dict = defaultdict(dict)
    with open(path, "r") as in_file:
        for line in in_file:
            e_id, prop, value = _get_cleaned_split(line, delimiter)
            ent_attr_dict[e_id][prop] = value
    return ent_attr_dict


def read_rel_triples(path: str, delimiter="\t") -> Dict[str, Dict[str, str]]:
    """Read relation triples.

    This functions returns the triples as dictionary. Containing
    the relations from left to right,i.e.  given a triple (s,p,o)
    the dictionary would be {s: {o: p}}

    Parameters
    ----------
    path : str
        Path to the file
    delimiter: str, default = tab
        Delimiter of the csv file

    Returns
    -------
    left_to_right: Dict[str, Dict[str, str]]
        Dictionary containing relation triples with subjects as key
        of outer dict
    """
    rel_dict = defaultdict(dict)
    with open(path, "r") as in_file:
        for line in in_file:
            left_id, rel, right_id = _get_cleaned_split(line, delimiter)
            rel_dict[left_id][right_id] = rel
    return rel_dict


def _read_two_column(path, delimiter="\t") -> ClusterHelper:
    links = {}
    with open(path, "r") as infile:
        for line in infile:
            left, right = _get_cleaned_split(line, delimiter)
            links[left] = right
    return ClusterHelper(links)


def read_links(path: str) -> ClusterHelper:
    """Read entity links.

    Parameters
    ----------
    path : str
        inpath

    Returns
    -------
    ClusterHelper
        ClusterHelper instance containing all links

    """
    return _read_two_column(path)


def _get_kg_name_from_path(path: str):
    if "D_W" in path:
        return "DBpedia", "Wikidata"
    if "D_Y" in path:
        return "DBpedia", "Yago"
    if "EN_DE" in path:
        return "English-DBpedia", "German-DBpedia"
    if "EN_FR" in path:
        return "English-DBpedia", "French-DBpedia"


def create_kg(path: str, one_or_two: str, name: str) -> KG:
    """Create a KG object from open ea files given in path.

    Parameters
    ----------
    path : str
        path to open ea files of dataset pair
    one_or_two : str
        which KG to create (either "1" or "2")
    name : str
        name of KG

    Returns
    -------
    KG
        knowledge graph object
    """
    attr_trip = read_attr_triples(os.path.join(path, f"attr_triples_{one_or_two}"))
    rel_trip = read_rel_triples(os.path.join(path, f"rel_triples_{one_or_two}"))
    return KG(attr_trip, rel_trip, name)


def from_openea(path: str) -> ERTask:
    """Create ERTask object from open ea files.

    Parameters
    ----------
    path : str
        path to openea files of dataset pair

    Returns
    -------
    ERTask
        er_task object
    """
    kg_names = _get_kg_name_from_path(path)
    kg1 = create_kg(path, "1", kg_names[0])
    kg2 = create_kg(path, "2", kg_names[1])
    return ERTask(kgs=[kg1, kg2], clusters=read_links(os.path.join(path, "ent_links")))



"""Implementation of common entity resolution metrics."""
from typing import Dict

from forayer.knowledge_graph import ClusterHelper, ERTask


def _tp_fp_fn(res: ClusterHelper, gold: ClusterHelper, only_tp_fp=False):
    tp = 0
    fp = 0
    fn = 0
    for pairs in res.all_pairs():
        if pairs in gold:
            tp += 1
        else:
            fp += 1
    if only_tp_fp:
        return tp, fp
    for pairs in gold.all_pairs():
        if pairs not in res:
            fn += 1
    return tp, fp, fn


def match_quality(res: ClusterHelper, gold: ClusterHelper) -> Dict:
    """Calculate match quality metrics.

    Parameters
    ----------
    res : ClusterHelper
        Result that should be measured.
    gold : ClusterHelper
        Perfect matches.

    Returns
    -------
    Dict
        Dictionary with metrics "fmeasure","prec" and "rec"

    """
    if len(res) == 0:
        return {"fmeasure": 0.0, "prec": 1.0, "rec": 0.0}
    if len(gold) == 0:
        return {"fmeasure": 0.0, "prec": 0.0, "rec": 0.0}
    tp, fp, fn = _tp_fp_fn(res, gold)
    prec = tp / (tp + fp)
    rec = tp / (tp + fn)
    if prec + rec == 0:
        fm = 0
    else:
        fm = 2 * (prec * rec) / (prec + rec)
    return {"fmeasure": fm, "prec": prec, "rec": rec}


def p_e_ratio(blocks: ClusterHelper, task: ERTask) -> float:
    """Calculate Pair/Entity ratio.

    Parameters
    ----------
    blocks : ClusterHelper
        ClusterHelper with block assignment
    task : ERTask
        Related entity resolution task.

    Returns
    -------
    float
        Pair/Entity ratio

    Raises
    ------
    ZeroDivisionError
        If task has length of zero
    """
    return blocks.number_of_links / len(task)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forayer',
 'forayer.datasets',
 'forayer.evaluation',
 'forayer.exploration',
 'forayer.input_output',
 'forayer.knowledge_graph',
 'forayer.transformation',
 'forayer.utils']

package_data = \
{'': ['*']}

install_requires = \
['gensim>=4.0.1,<5.0.0',
 'importlib-metadata',
 'ipykernel>=6.2.0,<7.0.0',
 'joblib>=1.0.1,<2.0.0',
 'nbformat>=5.1.3,<6.0.0',
 'networkx>=2.6.2,<3.0.0',
 'pandas>=1.3.2,<2.0.0',
 'plotly>=5.2.1,<6.0.0',
 'rdflib>=6.0.0,<7.0.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'tqdm>=4.56.0,<5.0.0',
 'wget>=3.2,<4.0']

extras_require = \
{'docs': ['sphinx>=4.0.2,<5.0.0', 'insegel>=1.1.0,<2.0.0']}

setup_kwargs = {
    'name': 'forayer',
    'version': '0.3.1',
    'description': 'First aid utilies for knowledge graph exploration with an entity centric approach',
    'long_description': '<p align="center">\n<img src="https://github.com/dobraczka/forayer/raw/main/docs/forayerlogo.png" alt="forayer logo", width=200/>\n</p>\n\n<h2 align="center"> forayer</h2>\n\n<p align="center">\n<a href="https://github.com/dobraczka/forayer/actions/workflows/main.yml"><img alt="Tests" src="https://github.com/dobraczka/forayer/actions/workflows/tests.yml/badge.svg?branch=main"></a>\n<a href="https://github.com/dobraczka/forayer/actions/workflows/quality.yml"><img alt="Linting" src="https://github.com/dobraczka/forayer/actions/workflows/quality.yml/badge.svg?branch=main"></a>\n<a><img alt="Test coverage" src="https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/dobraczka/6d07d95e43929bcbf9d031c2c8f2015f/raw/forayer_test_gist.json"></a>\n<a href="https://pypi.org/project/forayer"/><img alt="Stable python versions" src="https://img.shields.io/pypi/pyversions/forayer"></a>\n<a href="https://github.com/dobraczka/forayer/blob/main/LICENSE"><img alt="MIT License" src="https://img.shields.io/badge/license-MIT-blue"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\nAbout\n=====\nForayer is a library of **f**irst aid utilities for kn**o**wledge g**r**aph explor**a**tion with an entit**y** c**e**ntric app**r**oach.\nIt is intended to make data integration of knowledge graphs easier. With entities as first class citizens forayer is a toolset to aid in knowledge graph exploration for data integration and specifically entity resolution.\n\nYou can easily load pre-existing entity resolution tasks:\n\n```python\n  >>> from forayer.datasets import OpenEADataset\n  >>> ds = OpenEADataset(ds_pair="D_W",size="15K",version=1)\n  >>> ds.er_task\n  ERTask({DBpedia: (# entities: 15000, # entities_with_rel: 15000, # rel: 13359,\n  # entities_with_attributes: 13782, # attributes: 13782, # attr_values: 24995),\n  Wikidata: (# entities: 15000, # entities_with_rel: 15000, # rel: 13554,\n  # entities_with_attributes: 14376, # attributes: 14376, # attr_values: 114107)},\n  ClusterHelper(# elements:30000, # clusters:15000))\n```\n\nThis entity resolution task holds 2 knowledge graphs and a cluster of known matches. You can search in knowledge graphs:\n\n```python\n  >>> ds.er_task["DBpedia"].search("Dorothea")\n  KG(entities={\'http://dbpedia.org/resource/E801200\': \n  {\'http://dbpedia.org/ontology/activeYearsStartYear\': \'"1948"^^<http://www.w3.org/2001/XMLSchema#gYear>\',\n  \'http://dbpedia.org/ontology/activeYearsEndYear\': \'"2008"^^<http://www.w3.org/2001/XMLSchema#gYear>\',\n  \'http://dbpedia.org/ontology/birthName\': \'Dorothea Carothers Allen\',\n  \'http://dbpedia.org/ontology/alias\': \'Allen, Dorothea Carothers\',\n  \'http://dbpedia.org/ontology/birthYear\': \'"1923"^^<http://www.w3.org/2001/XMLSchema#gYear>\',\n  \'http://purl.org/dc/elements/1.1/description\': \'Film editor\',\n  \'http://dbpedia.org/ontology/birthDate\': \'"1923-12-03"^^<http://www.w3.org/2001/XMLSchema#date>\',\n  \'http://dbpedia.org/ontology/deathDate\': \'"2010-04-17"^^<http://www.w3.org/2001/XMLSchema#date>\', \n  \'http://dbpedia.org/ontology/deathYear\': \'"2010"^^<http://www.w3.org/2001/XMLSchema#gYear>\'}}, rel={}, name=DBpedia)\n```\n\nDecide to work with a smaller snippet of the resolution task:\n\n```python\n  >>> ert_sample = ds.er_task.sample(100)\n  >>> ert_sample\n  ERTask({DBpedia: (# entities: 100, # entities_with_rel: 6, # rel: 4,\n  # entities_with_attributes: 99, # attributes: 99, # attr_values: 274),\n  Wikidata: (# entities: 100, # entities_with_rel: 4, # rel: 4,\n  # entities_with_attributes: 100, # attributes: 100, # attr_values: 797)},\n  ClusterHelper(# elements:200, # clusters:100))\n```\n\nAnd much more can be found in the [user guide](https://forayer.readthedocs.io/en/latest/source/user_guide.html).\n\nInstallation\n============\n\nYou can install forayer via pip:\n\n```bash\n  pip install forayer\n```\n',
    'author': 'Daniel Obraczka',
    'author_email': 'obraczka@informatik.uni-leipzig.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dobraczka/forayer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""Knowledge graph and entity resolution related methods."""

from forayer.knowledge_graph.clusters import ClusterHelper
from forayer.knowledge_graph.er_task import ERTask
from forayer.knowledge_graph.kg import KG, AttributeEmbeddedKG

__all__ = ["ClusterHelper", "KG", "AttributeEmbeddedKG", "ERTask"]



"""Module concerned with utilizing word embeddings."""
import logging
import os
from typing import Any, Callable, Dict, List
from zipfile import ZipFile

import numpy as np
import wget
from gensim.models import KeyedVectors
from gensim.models.fasttext import load_facebook_model
from gensim.utils import tokenize

_EMBEDDING_INFO = {
    "fasttext": (
        "https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.simple.zip",
        "wiki.simple.zip",  # zip name
        "wiki.simple.bin",  # file containing wanted embeddings
    ),
    "glove": (
        "http://nlp.stanford.edu/data/glove.6B.zip",
        "glove.6B.zip",
        "glove.6B.300d.txt",
    ),
}


class AttributeVectorizer:
    """Vectorizer class to get attribute embeddings of entities with pre-trained embeddings.

    Attributes
    ----------
    tokenizer: Callable
        callable that tokenizes a string
    embedding_type: str
        type of pretrained embeddings
    vectors_path: str
        path to pre-trained embeddings
    wv: gensim.models.KeyedVectors
        word embeddings
    """

    def __init__(
        self,
        tokenizer: Callable = None,
        embedding_type: str = "glove",
        vectors_path: str = None,
        default_download_dir: str = None,
    ):
        """Initialize an AttributeVectorizer object and load the pre-trained embeddings.

        Parameters
        ----------
        tokenizer: Callable
            callable that tokenizes a string
        embedding_type: str
            type of pretrained embeddings
        vectors_path: str
            path to pre-trained embeddings
        default_download_dir: str
            directory where embeddings are downloaded if they are not present
            default is "./data/word_embeddings/"

        Raises
        ------
        TypeError
            if tokenizer is not callable
        ValueError
            if embedding_type is unknown or vectors_path does not exist
        """
        if tokenizer is None:
            self.tokenizer = tokenize
        else:
            if not hasattr(tokenizer, "___call__"):
                raise TypeError(
                    f"tokenizer should be a function, but was {type(tokenizer)}"
                )
            else:
                self.tokenizer = tokenizer
        embedding_type = embedding_type.lower()
        if embedding_type not in _EMBEDDING_INFO:
            raise ValueError(
                f"embedding_type has to be one of {set(_EMBEDDING_INFO.keys())}, not"
                f" {embedding_type}"
            )
        self.embedding_type = embedding_type
        if vectors_path is not None and not os.path.exists(vectors_path):
            raise ValueError(f"vectors_path: {vectors_path} does not exist")
        self.vectors_path = vectors_path
        self.default_download_dir = (
            default_download_dir
            if default_download_dir is not None
            else os.path.join("data", "word_embeddings")
        )
        self._download_embeddings_if_needed()
        self.wv = self._load_embeddings()
        self.vocab = {}
        self.seen_tokens = 0
        self.ignored_tokens = 0

    def reset_token_count(self):
        """Reset .seen_tokens and .ignored_tokens."""
        self.seen_tokens = 0
        self.ignored_tokens = 0

    def _download_embeddings_if_needed(self):
        if self.vectors_path is None:
            embeddings_dir = os.path.join(
                self.default_download_dir, self.embedding_type
            )
            if not os.path.exists(embeddings_dir):
                os.makedirs(embeddings_dir)
            dl_url, zip_name, embedding_file = _EMBEDDING_INFO[self.embedding_type]
            logging.info(
                f"Downloading {self.embedding_type} embeddings to {embeddings_dir}"
            )
            wget.download(dl_url, embeddings_dir)
            with ZipFile(os.path.join(embeddings_dir, zip_name), "r") as zip_obj:
                zip_obj.extractall(embeddings_dir)
            os.remove(os.path.join(embeddings_dir, zip_name))
            self.vectors_path = os.path.join(embeddings_dir, embedding_file)

    def _load_embeddings(self):
        logging.info("Loading word embeddings")
        if self.embedding_type == "fasttext":
            return load_facebook_model(self.vectors_path).wv
        try:
            return KeyedVectors.load_word2vec_format(self.vectors_path, binary=False)
        except ValueError:
            if self.embedding_type == "glove":
                wv = KeyedVectors.load_word2vec_format(
                    self.vectors_path, binary=False, no_header=True
                )
                return wv

    def vectorize(self, sentence: str) -> List[np.ndarray]:
        """Tokenize and vectorize a sentence with the given word embeddings.

        Parameters
        ----------
        sentence : str
            sentence to vectorize

        Returns
        -------
        vectorized: List[np.ndarray]
            List of token embeddings,

        Notes
        -----
        Ignores tokens that are not contained in the used embeddings
        Ignored tokens will be set to np.NaN
        """
        vectorized = []
        _number_of_tokens = 0  # initialize before in case loop is not executed
        for _number_of_tokens, word in enumerate(self.tokenizer(sentence), start=1):
            if word in self.wv:
                vectorized.append(self.wv[word])
            else:
                vectorized.append(np.NaN)
                self.ignored_tokens += 1
                logging.debug(f"Did not find embedding for {word}, ignoring")
        self.seen_tokens += _number_of_tokens
        return vectorized

    def vectorize_entity_attributes(
        self, attributes: Dict[Any, Any]
    ) -> Dict[Any, List[np.ndarray]]:
        """Tokenize and vectorize values of entity attributes.

        Parameters
        ----------
        attributes : Dict[Any, Any]
            dictionary of entity attributes with attribute names as keys

        Returns
        -------
        embedded_entity_attributes: Dict[Any, List[np.ndarray]]
            entity dicts with attribute values replaced with list of token embeddings
        """
        return {
            attr_name: self.vectorize(attr_value)
            for attr_name, attr_value in attributes.items()
        }



"""Methods to deal with entity clusters."""

import random
from itertools import chain, combinations
from typing import Any, Dict, Iterable, List, Set, Tuple, Union

from forayer.knowledge_graph.graph_based_clustering import (
    connected_components_from_container,
)
from forayer.utils.random_help import random_generator


class ClusterHelper:
    """Convenience class for entity clusters.

    The :class:`ClusterHelper` class holds a dict mapping entities to the respective cluster_id
    and a dict with cluster_id mapping to entity sets.
    The :meth:`.add()` and :meth:`.remove()` keep the respective dicts in sync.

    Attributes
    ----------
    clusters: Dict[str,Set[str]]
        maps cluster id to entity set
    entities: Dict[str,int]
        maps entity id to cluster id

    Examples
    --------
    >>> from forayer.knowledge_graph import ClusterHelper
    >>> ch = ClusterHelper([{"a1", "b1"}, {"a2", "b2"}])
    >>> print(ch)
    {0: {'a1', 'b1'}, 1: {'a2', 'b2'}}

    Add an element to a cluster

    >>> ch.add_to_cluster(0, "c1")
    >>> print(ch)
    {0: {'a1', 'b1', 'c1'}, 1: {'a2', 'b2'}}

    Add a new cluster

    >>> ch.add({"e2", "f1", "c3"})
    >>> print(ch)
    {0: {'a1', 'b1', 'c1'}, 1: {'a2', 'b2'}, 2: {'f1', 'e2', 'c3'}}

    Remove an element from a cluster

    >>> ch.remove("b1")
    >>> print(ch)
    {0: {'a1', 'c1'}, 1: {'a2', 'b2'}, 2: {'f1', 'e2', 'c3'}}

    The __contains__ function is smartly overloaded. You can
    check if an entity is in the ClusterHelper

    >>> "a1" in ch
    True

    If a cluster is present

    >>> {"c1","a1"} in ch
    True

    And even if a link exists or not

    >>> ("f1","e2") in ch
    True
    >>> ("a1","e2") in ch
    False

    To know the cluster id of an entity you can look it up with

    >>> ch.elements["a1"]
    0

    To get members of a cluster either use

    >>> ch.members(0)
    {'a1', 'b1', 'c1'}

    or simply

    >>> ch[0]
    {'a1', 'b1', 'c1'}

    """

    def _contains_overlaps(self, data):
        if len(data) > 1 and len(list(chain(*data))) > len(set.union(*data)):
            return True
        return False

    def _from_set_list(self, data: List[Set]):
        # check if contains overlaps
        if self._contains_overlaps(data):
            # merge overlapping
            data = connected_components_from_container(data)

        for cluster_id, inner in enumerate(data):
            if not isinstance(inner, set):
                raise TypeError(
                    f"Only set is allowed as list element, but got {type(inner)}"
                )
            inner_set = set()
            for inner_element in inner:
                inner_element = inner_element
                self.elements[inner_element] = cluster_id
                inner_set.add(inner_element)
            self.clusters[cluster_id] = inner_set

    def _from_dict(self, data: Dict):
        for cluster_id, dict_items in enumerate(data.items()):
            left, right = dict_items
            if left == right:
                raise ValueError(f"No selflinks allowed: {left} -> {right}")
            self.elements[left] = cluster_id
            self.clusters[cluster_id] = {left}
            self.elements[right] = cluster_id
            self.clusters[cluster_id].add(right)

    def _from_clusters(self, data: Dict):
        if self._contains_overlaps(data.values()):
            raise ValueError(
                "Entries with multiple memberships are not allowed, when specifying"
                " clusters and ids explicitly"
            )
        self.elements = {
            e_id: cluster_id for cluster_id, cluster in data.items() for e_id in cluster
        }
        self.clusters = data

    def __init__(
        self,
        data: Union[List[Set], Dict] = None,
    ):
        """Initialize a ClusterHelper object with clusters.

        Parameters
        ----------
        data : Union[List[Set], Dict]
            Clusters either as list of sets, or dict with
            links as key, value pairs, or dict with cluster id and set of members

        Raises
        ------
        TypeError
            if data is not dict or list
        ValueError
            For dict[cluster_id,member_set] if overlaps between clusters

        Notes
        -----
        Will try to merge clusters transitively if necessary.
        """
        self.elements = {}
        self.clusters = {}
        if data is None:
            return
        if not isinstance(data, (dict, list)):
            raise TypeError(f"Only list or dict allowed, but got {type(data)}")
        elif isinstance(data, list):
            self._from_set_list(data)
        elif isinstance(data, dict):
            if isinstance(list(data.values())[0], set):
                self._from_clusters(data)
            else:
                self._from_dict(data)

    def __repr__(self):
        return f"ClusterHelper(elements={str(self.elements)},clusters={str(self.clusters)})"

    def __str__(self):
        return str(self.clusters)

    def info(self):
        """Print general information about this object.

        Returns
        -------
        str
            information about number of entities and clusters
        """
        num_elements = len(self.elements)
        num_clusters = len(self.clusters)
        return (
            self.__class__.__name__
            + f"(# elements:{num_elements}, # clusters:{num_clusters})"
        )

    def links(self, key: str, always_return_set=False) -> Union[str, Set[str]]:
        """Get entities linked to this entity.

        Parameters
        ----------
        key : str
            entity id

        always_return_set: str
            If True, return set even if only one entity is contained
        Returns
        -------
        Union[str, Set[str]]
            Either the id of the single linked entity or a set of
            ids if there is more than one link
            If always_return_set is True, will always return set
        """
        cluster = self.clusters[self.elements[key]]
        other_members = cluster.difference({key})
        if not always_return_set and len(other_members) == 1:
            return next(iter(other_members))
        return other_members

    def all_pairs(self, key=None) -> Iterable[Tuple[Any, Any]]:
        """Get all entity pairs of a specific cluster or of all clusters.

        Parameters
        ----------
        key
            Cluster id.  If None, provides pairs of all clusters.

        Returns
        -------
        Generator[Tuple[Any, Any]]
            Generator that produces the wanted pairs.
        """
        if key is not None:
            return combinations(self.clusters[key], 2)
        # get pair combinations of clusters and chain generators
        return chain(*[combinations(cluster, 2) for cluster in self.clusters.values()])

    def members(self, key) -> Set:
        """Get members of a cluster.

        Parameters
        ----------
        key
            cluster id

        Returns
        -------
        Set
            cluster members
        """
        return self.clusters[key]

    def __getitem__(self, key):
        """Get cluster members.

        Parameters
        ----------
        key
            Cluster id

        Returns
        -------
        Set
            Cluster members
        """
        return self.clusters[key]

    def get(self, key, value=None):
        """Return cluster's element or default value.

        Tries to return the cluster with the cluster id == key.
        If None is found return provided value.

        Parameters
        ----------
        key
            Searched cluster id.
        value
            Default value to return in case id is not present.

        Returns
        -------
        Set
            Cluster with provided cluster_id.
        """
        try:
            return self[key]
        except KeyError:
            return value

    def __contains__(self, key):
        """Check if entities/links/clusters are contained.

        Parameters
        ----------
        key
            Either entity id, Tuple with two entities to check for a link
            between entities, or a clusters as set of entity ids
        """
        if isinstance(key, Set):
            return key in self.clusters.values()
        elif (
            isinstance(key, Tuple)
            and len(key) == 2
            and key[0] in self.elements
            and key[1] in self.elements
        ):
            return self.elements[key[0]] == self.elements[key[1]]
        else:
            return key in self.elements

    def __setitem__(self, key, value):
        """Not Implemented."""
        raise TypeError(
            "'ClusterHelper' does not support item assignment use .add() or .remove()"
        )

    def sample(self, n: int, seed: Union[int, random.Random] = None):
        """Sample n clusters.

        Parameters
        ----------
        n : int
            Number of clusters to return.
        seed : Union[int, random.Random]
            Seed for randomness or seeded random.Random object.
            Default is None.

        Returns
        -------
        ClusterHelper
           ClusterHelper with n clusters.
        """
        r_gen = random_generator(seed)
        return ClusterHelper(
            {c_id: cluster for c_id, cluster in r_gen.sample(self.clusters.items(), n)}
        )

    def merge(self, c1, c2, new_id=None):
        """Merge two clusters.

        Parameters
        ----------
        c1
            Id of one cluster to merge
        c2
            Id of other cluster to merge
        new_id
            New id of cluster, if None take c1

        Raises
        ------
        ValueError
            If cluster id(s) do not exist
        """
        if c1 not in self.clusters or c2 not in self.clusters:
            raise ValueError("Can only merge on existing cluster ids")
        cluster1 = self[c1]
        cluster2 = self[c2]
        if new_id:
            del self.clusters[c1]
            for e1 in cluster1:
                del self.elements[e1]
            self.add(cluster1, c_id=new_id)
        else:
            new_id = c1
        del self.clusters[c2]
        for e2 in cluster2:
            del self.elements[e2]
            self.add_to_cluster(new_id, e2)

    def add_link(self, e1, e2):
        """Add a new entity link.

        Either adds a link to an existing entity or
        creates a new cluster with both.

        Parameters
        ----------
        e1
           Id of one entity that will be linked
        e2
            Id of other entity that will be linked

        Returns
        ------
        c_id
            Id of cluster that was created, or
            of existing cluster that was enhanced
            Returns False if link already was present
        """
        if e1 not in self and e2 not in self:
            return self.add({e1, e2})
        elif e1 in self.elements and e2 in self.elements:
            if (e1, e2) in self:
                return False
            # merging
            cluster_id_1 = self.elements[e1]
            cluster_id_2 = self.elements[e2]
            return self.merge(cluster_id_1, cluster_id_2)
        elif e1 in self.elements:
            cluster_id = self.elements[e1]
            new_entity = e2
        elif e2 in self.elements:
            cluster_id = self.elements[e2]
            new_entity = e1
        self.clusters[cluster_id].add(new_entity)
        self.elements[new_entity] = cluster_id

    def add_to_cluster(self, c_id, new_entity):
        """Add an entity to a cluster.

        Parameters
        ----------
        c_id
            Cluster id where entity will be added
        new_entity
            Id of new entity

        Raises
        ------
        KeyError
            If cluster id unknonw
        ValueError
            If entity already belongs to other cluster
        """
        if c_id not in self.clusters:
            raise KeyError("Cluster id {c_id} unknown")
        if new_entity in self.elements:
            raise ValueError(
                "Entity id {new_entity} already belongs to {self.clusters[c_id]}"
            )
        self.elements[new_entity] = c_id
        self.clusters[c_id].add(new_entity)

    def add(
        self,
        new_entry: Set,
        c_id=None,
    ):
        """Add a new cluster.

        Parameters
        ----------
        new_entry : Set
            New cluster as set

        c_id
           Cluster id that will be assigned.
           If None, the next largest cluster id will be assigned
           Assuming cluster ids are integers

        Raises
        ------
        ValueError
            If entity id already present in other cluster
            Or if new cluster id cannot be inferred automatically
            by incrementing
        """
        if not isinstance(new_entry, set):
            raise TypeError(f"Expected set, but got {type(new_entry)}")
        else:
            if not len(new_entry.intersection(self.elements.keys())) == 0:
                raise ValueError("Set contains already present entries")
            else:
                max_cid = max(self.clusters.keys())
                if not isinstance(max_cid, int):
                    raise ValueError(
                        "Cluster Id cannot be automatically incremented, please provide"
                        " it explicitly"
                    )
                cluster_id = max_cid + 1
                self.clusters[cluster_id] = set()
                for e in new_entry:
                    self.elements[e] = cluster_id
                    self.clusters[cluster_id].add(e)

    def remove(self, entry: str):
        """Remove an entity.

        Parameters
        ----------
        entry : str
            entity to remove
        """
        cluster_id = self.elements[entry]
        del self.elements[entry]
        cluster = self.clusters[cluster_id]
        if len(cluster) == 2:
            for member in cluster:
                if member != entry:
                    del self.elements[member]
            del self.clusters[cluster_id]
        else:
            self.clusters[cluster_id].remove(entry)

    def remove_cluster(self, cluster_id):
        """Remove an entire cluster with the given cluster id.

        Parameters
        ----------
        cluster_id
            id of the cluster to remove
        """
        cluster_elements = self.clusters[cluster_id]
        for e in iter(cluster_elements):
            del self.elements[e]
        del self.clusters[cluster_id]

    def __eq__(self, other):
        if isinstance(other, ClusterHelper):
            return (self.clusters == other.clusters) and (
                self.elements == other.elements
            )
        return False

    def __len__(self):
        return len(self.clusters)

    @property
    def number_of_links(self):
        """Return the total number of links."""

        def number_of_pairs_in_set(s):
            n = len(s)
            return int(n * (n - 1) / 2)

        return sum(map(number_of_pairs_in_set, self.clusters.values()))



from __future__ import annotations

import random
from functools import reduce
from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, List, Set, Union

from forayer.utils.dict_help import dict_merge
from forayer.utils.random_help import random_generator

# avoid circular imports see: https://www.stefaanlippens.net/circular-imports-type-hints-python.html
if TYPE_CHECKING:

    from forayer.knowledge_graph import KG, ClusterHelper


class ERTask:
    """Class to model entity resolution task on knowledge graphs.

    Attributes
    ----------
    kgs_dict: Dict[str, KG]
        dictionary of given KGs, with KG names as keys
        KGs without names have their list index as key
    clusters: ClusterHelper
        known entity clusters
    """

    def __init__(self, kgs: List[KG], clusters: ClusterHelper = None):
        """Initialize an ERTask object.

        Parameters
        ----------
        kgs : List[KG]
            list of KGs that are to be integrated
        clusters : ClusterHelper
            known entity clusters
        """
        kgs_dict = {}
        for cur_id, k in enumerate(kgs):
            if k.name is None:
                k.name = str(cur_id)
            kgs_dict[k.name] = k
        self.kgs = kgs_dict
        self.clusters = clusters
        self.__inv_attr = None

    def __repr__(self):
        kg_info = "{" + ",".join([k.info() for _, k in self.kgs.items()]) + "}"
        return self.__class__.__name__ + f"({kg_info},{str(self.clusters.info())})"

    def __getitem__(self, key):
        return self.kgs[key]

    def sample(
        self, n: int, seed: Union[int, random.Random] = None, unmatched: int = None
    ):
        """Create a sample of the ERTask.

        Takes n clusters and creates the respective subgraphs.
        If unmatched is provided adds a number of entities without
        match to the subgraphs.

        Parameters
        ----------
        n : int
            Number of clusters.
        seed : Union[int, random.Random]
            Seed for randomness or seeded random.Random object.
            Default is None.
        unmatched : int
            Number of unmatched entities to include. Default is None.

        Returns
        -------
        ERTask
            downsampled ERTask

        Examples
        --------
        >>> from forayer.datasets import OpenEADataset
        >>> ds = OpenEADataset(ds_pair="D_W",size="15K",version=1)
        >>> ds.er_task.sample(n=10,unmatched=20)
        ERTask({DBpedia: (# entities: 26, # entities_with_rel: 0, # rel: 0, # entities_with_attributes: 26, # attributes: 26, # attr_values: 89),Wikidata: (# entities: 14, # entities_with_rel: 0, # rel: 0, # entities_with_attributes: 14, # attributes: 14, # attr_values: 102)},ClusterHelper(# elements:20, # clusters:10))

            You can use a seed to control reproducibility


        >>> ds.er_task.sample(n=10,seed=13,unmatched=20)
        ERTask({DBpedia: (# entities: 26, # entities_with_rel: 0, # rel: 0, # entities_with_attributes: 26, # attributes: 26, # attr_values: 93),Wikidata: (# entities: 14, # entities_with_rel: 0, # rel: 0, # entities_with_attributes: 14, # attributes: 14, # attr_values: 179)},ClusterHelper(# elements:20, # clusters:10))
        """
        r_gen = random_generator(seed)
        sample_clusters = self.clusters.sample(n, seed=r_gen)
        entity_ids = list(sample_clusters.elements.keys())
        if unmatched is not None:
            unm_ent = set()
            no_match_entities = self.without_match()
            if len(no_match_entities) >= n:
                unm_ent = r_gen.sample(no_match_entities)
            else:
                unm_ent = no_match_entities
                for cand in self.entity_ids:
                    if len(unm_ent) == unmatched:
                        break
                    if cand not in self.clusters:
                        unmatched.add(cand)
                    elif cand not in entity_ids:
                        cand_links = self.clusters.links(cand, always_return_set=True)
                        if not any(c in entity_ids or c in unm_ent for c in cand_links):
                            unm_ent.append(cand)
            entity_ids.extend(list(unm_ent))
        for _, k in self.kgs.items():
            sampled_kgs = [k.subgraph(entity_ids) for k_name, k in self.kgs.items()]
        return ERTask(kgs=sampled_kgs, clusters=sample_clusters)

    @property
    def entity_ids(self) -> Set[str]:
        """Return entity ids of all knowledge graphs.

        Returns
        -------
        Set[str]
            Entity ids of all knowledge graphs as set.
        """
        return set(chain(*[k.entity_ids for k in self.kgs.values()]))

    def all_entities(self, ignore_only_relational: bool = False) -> Dict[str, Dict]:
        """Return all entities.

        Parameters
        ----------
        ignore_only_relational : bool
            If True, ignores entities that only show up in the relations
            (and not in the entities with attributes)
        Returns
        -------
        Dict[str, Dict]
            all entities
        """
        all_attr_ent = reduce(dict_merge, [k.entities for k in self.kgs.values()])
        if not ignore_only_relational:
            for kg in self.kgs.values():
                for e in kg.rel:
                    if e not in all_attr_ent:
                        all_attr_ent[e] = {}
                for e in kg._inv_rel:
                    if e not in all_attr_ent:
                        all_attr_ent[e] = {}
        return all_attr_ent

    def without_match(self):
        """Return ids of entities without matches in given gold standard."""
        return [e for e in self.entity_ids if e not in self.clusters]

    def __len__(self):
        return sum([len(k) for k in self.kgs.values()])

    def __eq__(self, other):
        if isinstance(other, ERTask):
            return self.clusters == other.clusters and self.kgs == other.kgs
        return False

    def inverse_attr_dict(self) -> Dict[Any, Dict[str, str]]:
        """Create an attributes dictionary with unique attribute values as key.

        Returns
        -------
        Dict[Any, Dict[str,str]]
            inverse attribute dict
        """
        if self.__inv_attr is None:
            attr_to_kg_to_attr_name_to_ent = {}
            for kg_name, kg in self.kgs.items():
                for ent_id, ent_attr_dict in kg.entities.items():
                    for attr_name, attr_value in ent_attr_dict.items():
                        if attr_value not in attr_to_kg_to_attr_name_to_ent:
                            attr_to_kg_to_attr_name_to_ent[attr_value] = {}
                        if kg_name not in attr_to_kg_to_attr_name_to_ent[attr_value]:
                            attr_to_kg_to_attr_name_to_ent[attr_value][kg_name] = {}
                        if (
                            attr_name
                            not in attr_to_kg_to_attr_name_to_ent[attr_value][kg_name]
                        ):
                            attr_to_kg_to_attr_name_to_ent[attr_value][kg_name][
                                attr_name
                            ] = []
                        attr_to_kg_to_attr_name_to_ent[attr_value][kg_name][
                            attr_name
                        ].append(ent_id)
            self.__inv_attr = attr_to_kg_to_attr_name_to_ent
        return self.__inv_attr
